'use strict';

require('reflect-metadata');
var child_process = require('child_process');
var fbWatchman = require('fb-watchman');
var inversify = require('inversify');
var tslib_1 = require('tslib');
var path = require('path');
var fs = require('fs');

const Bindings = {
    Cli: Symbol('Cli'),
    Config: Symbol('Config'),
    ConfigManager: Symbol('ConfigManager'),
    Process: Symbol('Process'),
    Require: Symbol('require'),
    Spawn: Symbol('spawn'),
    Sync: Symbol('Sync'),
    Terminal: Symbol('Terminal'),
    WatchmanClient: Symbol('WatchmanClient'),
    WatchmanProcessor: Symbol('WatchmanProcessor'),
};

const DEFAULT_ARGUMENTS = {
    config: '',
    init: false,
};
let CliImpl = class CliImpl {
    constructor(process) {
        this.process = process;
    }
    getArguments() {
        const userArgs = Object.assign({}, DEFAULT_ARGUMENTS);
        const args = this.process.argv;
        for (let i = 0; i < args.length; i++) {
            const arg = args[i];
            switch (arg) {
                case '-i':
                case '--init':
                    userArgs.init = true;
                    break;
                case '-c':
                case '--config':
                    userArgs.config = this.parseConfigFile(i, args);
                    break;
                default:
                    break;
            }
        }
        return userArgs;
    }
    parseConfigFile(i, args) {
        if (i + 1 < args.length) {
            return path.resolve(args[i + 1]);
        }
        return this.process.cwd() + '/.watchman-processor.config.js';
    }
};
CliImpl = tslib_1.__decorate([
    inversify.injectable(),
    tslib_1.__param(0, inversify.inject(Bindings.Process))
], CliImpl);

const INIT_MSG = 'The watchman-processor configuration does not exist. \n\n' +
    'Run "watchman-processor --init" to create an example configuration file.\n';
const DEFAULT_CONFIG = {
    controlWatchman: true,
    debug: false,
    emoji: true,
    maxFileLength: 100,
    rsyncCmd: 'rsync',
    shell: '/bin/sh',
    syncDelete: true,
};
const DEFAULT_SUBSCRIPTION = {
    ignoreFolders: [],
    type: 'rsync',
};
let ConfigManagerImpl = class ConfigManagerImpl {
    constructor(cli, require, process) {
        this.cli = cli;
        this.require = require;
        this.process = process;
        /* istanbul ignore next */
        const homeFolder = this.process.env[(this.process.platform === 'win32') ? 'USERPROFILE' : 'HOME'];
        const args = this.cli.getArguments();
        this.options = Object.assign({
            confFile: args.config || path.resolve(homeFolder + '/.watchman-processor.config.js'),
            exampleConfFile: path.resolve(__dirname + '/example/watchman-processor.config.js'),
        });
    }
    getConfig() {
        const { cachedConfig, options, require } = this;
        if (cachedConfig) {
            return cachedConfig;
        }
        try {
            const configFile = path.resolve(options.confFile);
            if (!fs.existsSync(configFile)) {
                const error = new Error(INIT_MSG);
                error.name = 'init';
                throw error;
            }
            const userConfig = require(configFile);
            const config = this.cachedConfig = Object.assign({}, DEFAULT_CONFIG, userConfig);
            /* istanbul ignore if */
            if (!config.subscriptions) {
                throw new Error('No subscriptions defined');
            }
            // set the default settings for subscription configs
            const subscriptions = Object.keys(config.subscriptions);
            for (const name of subscriptions) {
                config.subscriptions[name] = Object.assign({}, DEFAULT_SUBSCRIPTION, config.subscriptions[name]);
            }
            return config;
        }
        catch (e) {
            return e;
        }
    }
    createConfig(
    /* istanbul ignore next */
    confFile = this.options.confFile, 
    /* istanbul ignore next */
    exampleConfFile = this.options.exampleConfFile) {
        const { options, process } = this;
        return new Promise((done, reject) => {
            const reader = fs.createReadStream(exampleConfFile);
            const writer = fs.createWriteStream(confFile);
            reader.on('error', reject);
            writer.on('error', reject);
            writer.on('close', () => {
                process.stdout.write('Done.  "' + options.confFile + '" created.\n');
                done();
            });
            reader.pipe(writer);
        });
    }
};
ConfigManagerImpl = tslib_1.__decorate([
    inversify.injectable(),
    tslib_1.__param(0, inversify.inject(Bindings.Cli)),
    tslib_1.__param(1, inversify.inject(Bindings.Require)),
    tslib_1.__param(2, inversify.inject(Bindings.Process))
], ConfigManagerImpl);

let SyncImpl = class SyncImpl {
    constructor(config, terminal, spawn) {
        this.config = config;
        this.terminal = terminal;
        this.spawn = spawn;
        this.processes = new Set();
    }
    syncFiles(subConfig, filesNames) {
        const ignoreFolders = subConfig.ignoreFolders;
        // remove files that are in the ignore folders
        const files = (filesNames || []).filter(file => !exists(ignoreFolders, file));
        // if there are too many files,
        // it might just be better to let rsync figure out what needs to be synced
        if (files.length > 0 && files.length < this.config.maxFileLength) {
            return this._syncSpecificFiles(subConfig, files);
        }
        else {
            return this._syncAllFiles(subConfig);
        }
    }
    end() {
        const { processes, terminal } = this;
        /* istanbul ignore else */
        if (processes.size > 0) {
            terminal.debug(`sync: kill rsync`);
            processes.forEach(proc => {
                proc.kill();
            });
        }
    }
    _syncAllFiles(subConfig) {
        const { config } = this;
        const { destination, ignoreFolders, source } = subConfig;
        const excludes = (`--exclude '${ignoreFolders.join(`' --exclude '`)}'`).split(' ');
        const deleteArg = config.syncDelete ? ['--delete'] : [];
        const args = ['-avz'].concat(deleteArg, excludes, [source, destination]);
        return this._exec(args);
    }
    _syncSpecificFiles(subConfig, files) {
        files = getUniqueFileFolders(files).concat(files);
        const { config } = this;
        const { destination, source } = subConfig;
        const includes = (`--include '${files.join(`' --include '`)}'`).split(' ');
        const deleteArg = config.syncDelete ? ['--delete'] : [];
        const args = ['-avz'].concat(deleteArg, includes, ['--exclude', `'*'`, source, destination]);
        return this._exec(args);
    }
    _exec(args) {
        const { spawn, config, terminal } = this;
        const cmdAndArgs = config.rsyncCmd + ' ' + args.join(' ');
        terminal.debug(cmdAndArgs);
        return new Promise((resolve, reject) => {
            const child = spawn(config.shell, ['-c', cmdAndArgs]);
            const errBuffer = [];
            this.processes.add(child);
            const done = this._execDoneHandler(errBuffer, child, resolve, reject);
            child.on('exit', done);
            child.on('close', done);
            if (child.stderr) {
                child.stderr.on('data', this._execErrDataHandler(errBuffer, terminal));
            }
            if (child.stdout) {
                child.stdout.on('data', this._execDataHandler(terminal));
            }
            child.on('error', this._execErrorHandler(child, reject));
        });
    }
    _execDataHandler(terminal) {
        return (data) => {
            terminal.debug(toString(data));
        };
    }
    _execErrDataHandler(errBuffer, terminal) {
        return (data) => {
            data = toString(data);
            errBuffer.push(data);
            terminal.debug(data);
        };
    }
    _execDoneHandler(errBuffer, child, resolve, reject) {
        return (code) => {
            this.processes.delete(child);
            code > 0 ? reject(errBuffer.join('\n')) : resolve();
        };
    }
    _execErrorHandler(child, reject) {
        return (err) => {
            this.processes.delete(child);
            reject(err);
        };
    }
};
SyncImpl = tslib_1.__decorate([
    inversify.injectable(),
    tslib_1.__param(0, inversify.inject(Bindings.Config)),
    tslib_1.__param(1, inversify.inject(Bindings.Terminal)),
    tslib_1.__param(2, inversify.inject(Bindings.Spawn))
], SyncImpl);
/**
 * Convert Buffer to string
 */
function toString(data) {
    /* istanbul ignore next */
    return data instanceof Buffer ? data.toString('utf8') : data;
}
/**
 * Given a list of files, return a list of folders leading up to those file names.
 *  Example:
 *    Given: ['example/js/1.js', 'example/js/2/.js']
 *    Will Return: ['example/', 'example/js/', 'example/js/1.js', 'example/js/2/.js']
 */
function getUniqueFileFolders(files) {
    const folders = [];
    const seen = new Map();
    for (const file of files) {
        const folderParts = getFolderParts(file.split('/'), seen);
        folders.push(...folderParts);
    }
    return folders;
}
function getFolderParts(folderParts, seen) {
    const folders = [];
    let folderPartsSum = '';
    for (const folderPart of folderParts) {
        if (folderPartsSum.length > 0) {
            folderPartsSum += '/';
        }
        folderPartsSum += folderPart;
        if (!seen.has(folderPartsSum)) {
            seen.set(folderPartsSum, undefined);
            folders.push(folderPartsSum);
        }
    }
    return folders;
}
/*
 * Return whether a value starts with a searchString in the given values array
 */
function exists(values, stringToSearch) {
    for (const value of values) {
        if (startsWith(stringToSearch, value)) {
            return true;
        }
    }
    return false;
}
/**
 * Return whether a string starts with another string
 */
function startsWith(value, searchString) {
    return value.substr(0, searchString.length) === searchString;
}

const fgBlack = '\x1b[30m';
const bgRed = '\x1b[41m';
const bgGreen = '\x1b[42m';
const bgYellow = '\x1b[43m';
const bgWhite = '\x1b[47m';
let TerminalImpl = class TerminalImpl {
    constructor(config, process) {
        this.config = config;
        this.process = process;
    }
    error(err) {
        const msg = err.toString();
        this.process.stderr.write(`${msg}`);
        if (typeof err !== 'string') {
            this.process.stderr.write(`${err.stack}`);
        }
    }
    debug(msg) {
        if (this.config.debug && typeof msg === 'string' && msg.length > 0) {
            msg = msg.trim();
            this.process.stdout.write(msg + '\n');
        }
    }
    setState(configEntry, state, statusMessage) {
        configEntry.state = state;
        configEntry.statusMessage = statusMessage;
        this.render();
    }
    render() {
        /* istanbul ignore if */
        if (this._quiet()) {
            return;
        }
        this._clear();
        const subscriptions = Object.keys(this.config.subscriptions);
        const statusBuffer = [];
        for (const name of subscriptions) {
            const subscription = this.config.subscriptions[name];
            const { state, statusMessage } = subscription;
            if (statusMessage) {
                statusBuffer.push(statusMessage);
            }
            this._renderState(name, state);
        }
        this._log('\n', bgWhite);
        this._reset();
        this.error(statusBuffer.join('\n'));
    }
    _renderState(name, state) {
        const log = this._log.bind(this);
        const emoji = this._emoji.bind(this);
        if (state === 'good') {
            log(emoji('👍 ') + ' ' + name + ' ', bgGreen);
        }
        else if (state === 'running') {
            log(emoji('🏃 ') + ' ' + name + ' ', bgYellow);
        }
        else if (state === 'error') {
            log(emoji('💀 ') + ' ' + name + ' ', bgRed);
        }
        else {
            log(emoji('⚡️ ') + ' ' + name + ' ', bgWhite);
        }
    }
    _emoji(str) {
        /* istanbul ignore next */
        return this.config.emoji ? str : '';
    }
    _clear() {
        this.process.stdout.write('\x1b[H\x1b[J');
    }
    _reset() {
        this.process.stdout.write('\x1b[0m');
    }
    _log(msg, backgroundColor) {
        this.process.stdout.write(`${backgroundColor}${fgBlack}${msg}`);
        this._reset();
    }
    _quiet() {
        return (!this
            || !this.config
            || this.config.debug
            || !this.config.subscriptions);
    }
};
TerminalImpl = tslib_1.__decorate([
    inversify.injectable(),
    tslib_1.__param(0, inversify.inject(Bindings.Config)),
    tslib_1.__param(1, inversify.inject(Bindings.Process))
], TerminalImpl);

let WatchmanProcessorImpl = class WatchmanProcessorImpl {
    constructor(config, client, terminal, sync) {
        this.config = config;
        this.client = client;
        this.terminal = terminal;
        this.sync = sync;
    }
    start() {
        const { client, terminal } = this;
        terminal.debug('watchman: initialize');
        const onCapabilityCheck = this.onCapabilityCheck.bind(this);
        client.capabilityCheck({}, onCapabilityCheck);
    }
    end() {
        const { client, config, sync } = this;
        const subscriptions = config.subscriptions;
        const promises = [];
        for (const name of Object.keys(subscriptions)) {
            const sub = subscriptions[name];
            promises.push(this.unsubscribe(sub.source, name));
        }
        sync.end();
        return Promise.all(promises)
            .then(() => {
            if (config.controlWatchman) {
                this.shutdown();
            }
            else {
                client.end();
            }
        });
    }
    onCapabilityCheck(error) {
        const terminal = this.terminal;
        if (error) {
            terminal.error(error);
            return;
        }
        terminal.render();
        const client = this.client;
        const onSubscription = this.onSubscription.bind(this);
        const promises = [];
        const subscriptions = Object.keys(this.config.subscriptions);
        for (const name of subscriptions) {
            const sub = this.config.subscriptions[name];
            const expression = sub.watchExpression || ['allof', ['type', 'f']];
            for (const folder of sub.ignoreFolders) {
                expression.push(['not', ['dirname', folder]]);
            }
            promises.push(this.subscribe(path.resolve(sub.source), name, expression));
        }
        const render = terminal.render.bind(terminal);
        const errHandler = terminal.error.bind(terminal);
        Promise.all(promises).then(render).catch(errHandler);
        // subscription is fired regardless of which subscriber fired it
        client.on('subscription', onSubscription);
    }
    onSubscription(resp) {
        const config = this.config;
        const subscription = resp && resp.subscription;
        const files = resp.files;
        const subConfig = config.subscriptions[subscription];
        this.syncFiles(subConfig, files);
    }
    syncFiles(subConfig, files) {
        const { sync, terminal } = this;
        terminal.setState(subConfig, 'running');
        const fileNames = (files || []).map(file => file.name);
        sync.syncFiles(subConfig, fileNames)
            .then(() => {
            terminal.setState(subConfig, 'good');
        })
            .catch(err => {
            terminal.setState(subConfig, 'error', err);
        });
    }
    subscribe(folder, name, expression) {
        const terminal = this.terminal;
        const client = this.client;
        const sub = {
            expression,
            fields: ['name', 'exists'],
            relative_root: '',
        };
        terminal.debug(`subscribe: ${name}`);
        return new Promise((resolve, reject) => {
            client.command(['subscribe', folder, name, sub], (error) => {
                error ? reject('failed to start: ' + error) : resolve();
            });
        });
    }
    unsubscribe(folder, name) {
        const terminal = this.terminal;
        const client = this.client;
        terminal.debug(`unsubscribe: ${name}`);
        return new Promise(resolve => {
            client.command(['unsubscribe', folder, name], resolve);
        });
    }
    shutdown() {
        const terminal = this.terminal;
        const client = this.client;
        terminal.debug(`watchman: shutdown`);
        return new Promise(resolve => {
            client.command(['shutdown-server'], resolve);
        });
    }
};
WatchmanProcessorImpl = tslib_1.__decorate([
    inversify.injectable(),
    tslib_1.__param(0, inversify.inject(Bindings.Config)),
    tslib_1.__param(1, inversify.inject(Bindings.WatchmanClient)),
    tslib_1.__param(2, inversify.inject(Bindings.Terminal)),
    tslib_1.__param(3, inversify.inject(Bindings.Sync))
], WatchmanProcessorImpl);

const container = new inversify.Container();
// setup core functionality and externals
container.bind(Bindings.Process).toConstantValue(process);
container.bind(Bindings.Spawn).toConstantValue(child_process.spawn);
container.bind(Bindings.Require).toConstantValue(require);
container.bind(Bindings.WatchmanClient).toConstantValue(new fbWatchman.Client());
// setup the main classes
container.bind(Bindings.Cli).to(CliImpl);
container.bind(Bindings.ConfigManager).to(ConfigManagerImpl);
container.bind(Bindings.Config).toDynamicValue(() => {
    return container.get(Bindings.ConfigManager).getConfig();
});
container.bind(Bindings.Terminal).to(TerminalImpl);
container.bind(Bindings.Sync).to(SyncImpl);
container.bind(Bindings.WatchmanProcessor).to(WatchmanProcessorImpl);

const cli = container.get(Bindings.Cli);
const configManager = container.get(Bindings.ConfigManager);
const watchmanProcessor = container.get(Bindings.WatchmanProcessor);
const args = cli.getArguments();
let processor = watchmanProcessor;
if (args.init) {
    configManager.createConfig();
    process.abort();
}
else {
    const config = configManager.getConfig();
    if (config instanceof Error) {
        processor = null;
        process.stderr.write(config.message + '\n');
        if (config.name !== 'init') {
            process.stderr.write(config.name);
            if (config.stack) {
                process.stderr.write(config.stack);
            }
        }
    }
}
var processor$1 = processor;

module.exports = processor$1;
//# sourceMappingURL=index.js.map
